@@ -14,7 +14,6 @@ import aiohttp.http
 import requests
 import requests.utils
 from requests.models import Response
-from voluptuous import MultipleInvalid
 
 from .version import __version__
 from .errors import (
@@ -26,19 +25,12 @@ from .errors import (
     PermissionRequiredError,
 )
 from .models import Factors, Insights, Score
-from .validation import validate_report, validate_transaction
+from .request import prepare_report, prepare_transaction
 
 
-_AIOHTTP_UA = "minFraud-API/%s %s" % (
-    __version__,
-    aiohttp.http.SERVER_SOFTWARE,
-)
-
-_REQUEST_UA = "minFraud-API/%s %s" % (
-    __version__,
-    requests.utils.default_user_agent(),
-)
+_AIOHTTP_UA = f"minFraud-API/{__version__} {aiohttp.http.SERVER_SOFTWARE}"
 
+_REQUEST_UA = f"minFraud-API/{__version__} {requests.utils.default_user_agent()}"
 
 # pylint: disable=too-many-instance-attributes, missing-class-docstring
 class BaseClient:
@@ -52,7 +44,7 @@ class BaseClient:
     _factors_uri: str
     _report_uri: str
 
-    def __init__(  # pylint: disable=too-many-arguments
+    def __init__(
         self,
         account_id: int,
         license_key: str,
@@ -65,65 +57,33 @@ class BaseClient:
         self._license_key = license_key
         self._timeout = timeout
 
-        base_uri = u"https://{0:s}/minfraud/v2.0".format(host)
+        base_uri = f"https://{host}/minfraud/v2.0"
         self._score_uri = "/".join([base_uri, "score"])
         self._insights_uri = "/".join([base_uri, "insights"])
         self._factors_uri = "/".join([base_uri, "factors"])
         self._report_uri = "/".join([base_uri, "transactions", "report"])
 
-    def _prepare_report(self, request: Dict[str, Any], validate: bool):
-        cleaned_request = self._copy_and_clean(request)
-        if validate:
-            try:
-                validate_report(cleaned_request)
-            except MultipleInvalid as ex:
-                raise InvalidRequestError("Invalid report data: {0}".format(ex)) from ex
-        return cleaned_request
-
-    def _prepare_transaction(self, request: Dict[str, Any], validate: bool):
-        cleaned_request = self._copy_and_clean(request)
-        if validate:
-            try:
-                validate_transaction(cleaned_request)
-            except MultipleInvalid as ex:
-                raise InvalidRequestError(
-                    "Invalid transaction data: {0}".format(ex)
-                ) from ex
-        return cleaned_request
-
-    def _copy_and_clean(self, data: Any) -> Any:
-        """Create a copy of the data structure with Nones removed."""
-        if isinstance(data, dict):
-            return dict(
-                (k, self._copy_and_clean(v)) for (k, v) in data.items() if v is not None
-            )
-        if isinstance(data, (list, set, tuple)):
-            return [self._copy_and_clean(x) for x in data if x is not None]
-        return data
-
     def _handle_success(
         self,
-        body: str,
+        raw_body: str,
         uri: str,
         model_class: Union[Type[Factors], Type[Score], Type[Insights]],
     ) -> Union[Score, Factors, Insights]:
         """Handle successful response."""
         try:
-            decoded_body = json.loads(body)
+            decoded_body = json.loads(raw_body)
         except ValueError as ex:
             raise MinFraudError(
-                "Received a 200 response"
-                " but could not decode the response as "
-                "JSON: {0}".format(body),
+                f"Received a 200 response but could not decode the response as JSON: {raw_body}",
                 200,
                 uri,
             ) from ex
-        if "ip_address" in body:
+        if "ip_address" in decoded_body:
             decoded_body["ip_address"]["_locales"] = self._locales
         return model_class(decoded_body)  # type: ignore
 
     def _exception_for_error(
-        self, status: int, content_type: str, body: str, uri: str
+        self, status: int, content_type: str, raw_body: str, uri: str
     ) -> Union[
         AuthenticationError,
         InsufficientFundsError,
@@ -134,13 +94,13 @@ class BaseClient:
         """Returns the exception for the error responses."""
 
         if 400 <= status < 500:
-            return self._exception_for_4xx_status(status, content_type, body, uri)
+            return self._exception_for_4xx_status(status, content_type, raw_body, uri)
         if 500 <= status < 600:
-            return self._exception_for_5xx_status(status, body, uri)
-        return self._exception_for_unexpected_status(status, body, uri)
+            return self._exception_for_5xx_status(status, raw_body, uri)
+        return self._exception_for_unexpected_status(status, raw_body, uri)
 
     def _exception_for_4xx_status(
-        self, status: int, content_type: str, body: str, uri: str
+        self, status: int, content_type: str, raw_body: str, uri: str
     ) -> Union[
         AuthenticationError,
         InsufficientFundsError,
@@ -149,28 +109,26 @@ class BaseClient:
         PermissionRequiredError,
     ]:
         """Returns exception for error responses with 4xx status codes."""
-        if not body:
+        if not raw_body:
             return HTTPError(
-                "Received a {0} error with no body".format(status), status, uri, body
+                f"Received a {status} error with no body", status, uri, raw_body
             )
         if content_type.find("json") == -1:
             return HTTPError(
-                "Received a {0} with the following " "body: {1}".format(status, body),
+                f"Received a {status} with the following body: {raw_body}",
                 status,
                 uri,
-                body,
+                raw_body,
             )
         try:
-            decoded_body = json.loads(body)
+            decoded_body = json.loads(raw_body)
         except ValueError:
             return HTTPError(
-                "Received a {status:d} error but it did not include"
-                " the expected JSON body: {content}".format(
-                    status=status, content=body
-                ),
+                f"Received a {status} error but it did not "
+                + f"include the expected JSON body: {raw_body}",
                 status,
                 uri,
-                body,
+                raw_body,
             )
         else:
             if "code" in decoded_body and "error" in decoded_body:
@@ -178,11 +136,11 @@ class BaseClient:
                     decoded_body.get("error"), decoded_body.get("code"), status, uri
                 )
             return HTTPError(
-                "Error response contains JSON but it does not specify code"
-                " or error keys: {0}".format(body),
+                "Error response contains JSON but it does not "
+                + f"specify code or error keys: {raw_body}",
                 status,
                 uri,
-                body,
+                raw_body,
             )
 
     @staticmethod
@@ -212,29 +170,29 @@ class BaseClient:
     @staticmethod
     def _exception_for_5xx_status(
         status: int,
-        body: Optional[str],
+        raw_body: Optional[str],
         uri: str,
     ) -> HTTPError:
         """Returns exception for error response with 5xx status codes."""
         return HTTPError(
-            u"Received a server error ({0}) for " u"{1}".format(status, uri),
+            f"Received a server error ({status}) for {uri}",
             status,
             uri,
-            body,
+            raw_body,
         )
 
     @staticmethod
     def _exception_for_unexpected_status(
         status: int,
-        body: Optional[str],
+        raw_body: Optional[str],
         uri: str,
     ) -> HTTPError:
         """Returns exception for responses with unexpected status codes."""
         return HTTPError(
-            u"Received an unexpected HTTP status " u"({0}) for {1}".format(status, uri),
+            f"Received an unexpected HTTP status ({status}) for {uri}",
             status,
             uri,
-            body,
+            raw_body,
         )
 
 
@@ -242,14 +200,16 @@ class AsyncClient(BaseClient):
     """Async client for accessing the minFraud web services."""
 
     _existing_session: aiohttp.ClientSession
+    _proxy: Optional[str]
 
-    def __init__(  # pylint: disable=too-many-arguments
+    def __init__(
         self,
         account_id: int,
         license_key: str,
         host: str = "minfraud.maxmind.com",
         locales: Tuple[str, ...] = ("en",),
         timeout: float = 60,
+        proxy: Optional[str] = None,
     ) -> None:
         """Constructor for AsyncClient.
 
@@ -261,30 +221,42 @@ class AsyncClient(BaseClient):
         :type host: str
         :param locales: A tuple of locale codes to use in name property
         :type locales: tuple[str]
-        :param timeout: The timeout in seconts to use when waiting on the request.
+        :param timeout: The timeout in seconds to use when waiting on the request.
           This sets both the connect timeout and the read timeout. The default is
           60.
         :type timeout: float
+        :param proxy: The URL of an HTTP proxy to use. It may optionally include
+          a basic auth username and password, e.g.,
+          ``http://username:password@host:port``.
         :return: Client object
         :rtype: Client
         """
         super().__init__(account_id, license_key, host, locales, timeout)
+        self._proxy = proxy
 
     async def factors(
-        self, transaction: Dict[str, Any], validate: bool = True
+        self,
+        transaction: Dict[str, Any],
+        validate: bool = True,
+        hash_email: bool = False,
     ) -> Factors:
         """Query Factors endpoint with transaction data.
 
         :param transaction: A dictionary containing the transaction to be
-          sent to the minFraud Insights web service as specified in the `REST
+          sent to the minFraud Factors web service as specified in the `REST
           API documentation
-          <https://dev.maxmind.com/minfraud/#Request_Body>`_.
+          <https://dev.maxmind.com/minfraud/api-documentation/requests?lang=en>`_.
         :type transaction: dict
         :param validate: If set to false, validation of the transaction
           dictionary will be disabled. This validation helps ensure that your
           request is correct before sending it to MaxMind. Validation raises an
           InvalidRequestError.
         :type validate: bool
+        :param hash_email: By default, the email address is sent in plain text.
+          If this is set to ``True``, the email address will be normalized and
+          converted to an MD5 hash before the request is sent. The email domain
+          will continue to be sent in plain text.
+        :type hash_email: bool
         :return: A Factors model object
         :rtype: Factors
         :raises: AuthenticationError, InsufficientFundsError,
@@ -292,24 +264,38 @@ class AsyncClient(BaseClient):
         """
         return cast(
             Factors,
-            await self._response_for(self._factors_uri, Factors, transaction, validate),
+            await self._response_for(
+                self._factors_uri,
+                Factors,
+                transaction,
+                validate,
+                hash_email,
+            ),
         )
 
     async def insights(
-        self, transaction: Dict[str, Any], validate: bool = True
+        self,
+        transaction: Dict[str, Any],
+        validate: bool = True,
+        hash_email: bool = False,
     ) -> Insights:
         """Query Insights endpoint with transaction data.
 
         :param transaction: A dictionary containing the transaction to be
           sent to the minFraud Insights web service as specified in the `REST
           API documentation
-          <https://dev.maxmind.com/minfraud/#Request_Body>`_.
+          <https://dev.maxmind.com/minfraud/api-documentation/requests?lang=en>`_.
         :type transaction: dict
         :param validate: If set to false, validation of the transaction
           dictionary will be disabled. This validation helps ensure that your
           request is correct before sending it to MaxMind. Validation raises an
           InvalidRequestError.
         :type validate: bool
+        :param hash_email: By default, the email address is sent in plain text.
+          If this is set to ``True``, the email address will be normalized and
+          converted to an MD5 hash before the request is sent. The email domain
+          will continue to be sent in plain text.
+        :type hash_email: bool
         :return: An Insights model object
         :rtype: Insights
         :raises: AuthenticationError, InsufficientFundsError,
@@ -318,23 +304,37 @@ class AsyncClient(BaseClient):
         return cast(
             Insights,
             await self._response_for(
-                self._insights_uri, Insights, transaction, validate
+                self._insights_uri,
+                Insights,
+                transaction,
+                validate,
+                hash_email,
             ),
         )
 
-    async def score(self, transaction: Dict[str, Any], validate: bool = True) -> Score:
+    async def score(
+        self,
+        transaction: Dict[str, Any],
+        validate: bool = True,
+        hash_email: bool = False,
+    ) -> Score:
         """Query Score endpoint with transaction data.
 
         :param transaction: A dictionary containing the transaction to be
           sent to the minFraud Score web service as specified in the `REST API
           documentation
-          <https://dev.maxmind.com/minfraud/#Request_Body>`_.
+          <https://dev.maxmind.com/minfraud/api-documentation/requests?lang=en>`_.
         :type transaction: dict
         :param validate: If set to false, validation of the transaction
           dictionary will be disabled. This validation helps ensure that your
           request is correct before sending it to MaxMind. Validation raises an
           InvalidRequestError.
         :type validate: bool
+        :param hash_email: By default, the email address is sent in plain text.
+          If this is set to ``True``, the email address will be normalized and
+          converted to an MD5 hash before the request is sent. The email domain
+          will continue to be sent in plain text.
+        :type hash_email: bool
         :return: A Score model object
         :rtype: Score
         :raises: AuthenticationError, InsufficientFundsError,
@@ -342,7 +342,13 @@ class AsyncClient(BaseClient):
         """
         return cast(
             Score,
-            await self._response_for(self._score_uri, Score, transaction, validate),
+            await self._response_for(
+                self._score_uri,
+                Score,
+                transaction,
+                validate,
+                hash_email,
+            ),
         )
 
     async def report(
@@ -353,7 +359,7 @@ class AsyncClient(BaseClient):
         :param report: A dictionary containing the transaction report to be sent
           to the Report Transations web service as specified in the `REST API`
           documentation
-          <https://dev.maxmind.com/minfraud/report-transaction/#Request_Body>_.
+          <https://dev.maxmind.com/minfraud/report-a-transaction?lang=en>_.
         :type report: dict
         :param validate: If set to false, validation of the report dictionary
           will be disabled. This validation helps ensure that your request is
@@ -365,15 +371,15 @@ class AsyncClient(BaseClient):
         :raises: AuthenticationError, InvalidRequestError, HTTPError,
           MinFraudError,
         """
-        prepared_request = self._prepare_report(report, validate)
+        prepared_request = prepare_report(report, validate)
         uri = self._report_uri
         async with await self._do_request(uri, prepared_request) as response:
             status = response.status
             content_type = response.content_type
-            body = await response.text()
+            raw_body = await response.text()
 
             if status != 204:
-                raise self._exception_for_error(status, content_type, body, uri)
+                raise self._exception_for_error(status, content_type, raw_body, uri)
 
     async def _response_for(
         self,
@@ -381,23 +387,24 @@ class AsyncClient(BaseClient):
         model_class: Union[Type[Factors], Type[Score], Type[Insights]],
         request: Dict[str, Any],
         validate: bool,
+        hash_email: bool,
     ) -> Union[Score, Factors, Insights]:
         """Send request and create response object."""
-        prepared_request = self._prepare_transaction(request, validate)
+        prepared_request = prepare_transaction(request, validate, hash_email)
         async with await self._do_request(uri, prepared_request) as response:
             status = response.status
             content_type = response.content_type
-            body = await response.text()
+            raw_body = await response.text()
 
             if status != 200:
-                raise self._exception_for_error(status, content_type, body, uri)
-            return self._handle_success(body, uri, model_class)
+                raise self._exception_for_error(status, content_type, raw_body, uri)
+            return self._handle_success(raw_body, uri, model_class)
 
     async def _do_request(
         self, uri: str, data: Dict[str, Any]
     ) -> aiohttp.ClientResponse:
         session = await self._session()
-        return await session.post(uri, json=data)
+        return await session.post(uri, json=data, proxy=self._proxy)
 
     async def _session(self) -> aiohttp.ClientSession:
         if not hasattr(self, "_existing_session"):
@@ -427,15 +434,17 @@ class AsyncClient(BaseClient):
 class Client(BaseClient):
     """Synchronous client for accessing the minFraud web services."""
 
+    _proxies: Optional[Dict[str, str]]
     _session: requests.Session
 
-    def __init__(  # pylint: disable=too-many-arguments
+    def __init__(
         self,
         account_id: int,
         license_key: str,
         host: str = "minfraud.maxmind.com",
         locales: Tuple[str, ...] = ("en",),
         timeout: float = 60,
+        proxy: Optional[str] = None,
     ) -> None:
         """Constructor for Client.
 
@@ -447,9 +456,12 @@ class Client(BaseClient):
         :type host: str
         :param locales: A tuple of locale codes to use in name property
         :type locales: tuple[str]
-        :param timeout: The timeout in seconts to use when waiting on the request.
+        :param timeout: The timeout in seconds to use when waiting on the request.
           This sets both the connect timeout and the read timeout. The default is
           60.
+        :param proxy: The URL of an HTTP proxy to use. It may optionally include
+          a basic auth username and password, e.g.,
+          ``http://username:password@host:port``.
         :type timeout: float
         :return: Client object
         :rtype: Client
@@ -461,19 +473,34 @@ class Client(BaseClient):
         self._session.headers["Accept"] = "application/json"
         self._session.headers["User-Agent"] = _REQUEST_UA
 
-    def factors(self, transaction: Dict[str, Any], validate: bool = True) -> Factors:
+        if proxy is None:
+            self._proxies = None
+        else:
+            self._proxies = {"https": proxy}
+
+    def factors(
+        self,
+        transaction: Dict[str, Any],
+        validate: bool = True,
+        hash_email: bool = False,
+    ) -> Factors:
         """Query Factors endpoint with transaction data.
 
         :param transaction: A dictionary containing the transaction to be
-          sent to the minFraud Insights web service as specified in the `REST
+          sent to the minFraud Factors web service as specified in the `REST
           API documentation
-          <https://dev.maxmind.com/minfraud/#Request_Body>`_.
+          <https://dev.maxmind.com/minfraud/api-documentation/requests?lang=en>`_.
         :type transaction: dict
         :param validate: If set to false, validation of the transaction
           dictionary will be disabled. This validation helps ensure that your
           request is correct before sending it to MaxMind. Validation raises an
           InvalidRequestError.
         :type validate: bool
+        :param hash_email: By default, the email address is sent in plain text.
+          If this is set to ``True``, the email address will be normalized and
+          converted to an MD5 hash before the request is sent. The email domain
+          will continue to be sent in plain text.
+        :type hash_email: bool
         :return: A Factors model object
         :rtype: Factors
         :raises: AuthenticationError, InsufficientFundsError,
@@ -481,22 +508,38 @@ class Client(BaseClient):
         """
         return cast(
             Factors,
-            self._response_for(self._factors_uri, Factors, transaction, validate),
+            self._response_for(
+                self._factors_uri,
+                Factors,
+                transaction,
+                validate,
+                hash_email,
+            ),
         )
 
-    def insights(self, transaction: Dict[str, Any], validate: bool = True) -> Insights:
+    def insights(
+        self,
+        transaction: Dict[str, Any],
+        validate: bool = True,
+        hash_email: bool = False,
+    ) -> Insights:
         """Query Insights endpoint with transaction data.
 
         :param transaction: A dictionary containing the transaction to be
           sent to the minFraud Insights web service as specified in the `REST
           API documentation
-          <https://dev.maxmind.com/minfraud/#Request_Body>`_.
+          <https://dev.maxmind.com/minfraud/api-documentation/requests?lang=en>`_.
         :type transaction: dict
         :param validate: If set to false, validation of the transaction
           dictionary will be disabled. This validation helps ensure that your
           request is correct before sending it to MaxMind. Validation raises an
           InvalidRequestError.
         :type validate: bool
+        :param hash_email: By default, the email address is sent in plain text.
+          If this is set to ``True``, the email address will be normalized and
+          converted to an MD5 hash before the request is sent. The email domain
+          will continue to be sent in plain text.
+        :type hash_email: bool
         :return: An Insights model object
         :rtype: Insights
         :raises: AuthenticationError, InsufficientFundsError,
@@ -504,29 +547,52 @@ class Client(BaseClient):
         """
         return cast(
             Insights,
-            self._response_for(self._insights_uri, Insights, transaction, validate),
+            self._response_for(
+                self._insights_uri,
+                Insights,
+                transaction,
+                validate,
+                hash_email,
+            ),
         )
 
-    def score(self, transaction: Dict[str, Any], validate: bool = True) -> Score:
+    def score(
+        self,
+        transaction: Dict[str, Any],
+        validate: bool = True,
+        hash_email: bool = False,
+    ) -> Score:
         """Query Score endpoint with transaction data.
 
         :param transaction: A dictionary containing the transaction to be
           sent to the minFraud Score web service as specified in the `REST API
           documentation
-          <https://dev.maxmind.com/minfraud/#Request_Body>`_.
+          <https://dev.maxmind.com/minfraud/api-documentation/requests?lang=en>`_.
         :type transaction: dict
         :param validate: If set to false, validation of the transaction
           dictionary will be disabled. This validation helps ensure that your
           request is correct before sending it to MaxMind. Validation raises an
           InvalidRequestError.
         :type validate: bool
+        :param hash_email: By default, the email address is sent in plain text.
+          If this is set to ``True``, the email address will be normalized and
+          converted to an MD5 hash before the request is sent. The email domain
+          will continue to be sent in plain text.
+        :type hash_email: bool
         :return: A Score model object
         :rtype: Score
         :raises: AuthenticationError, InsufficientFundsError,
           InvalidRequestError, HTTPError, MinFraudError,
         """
         return cast(
-            Score, self._response_for(self._score_uri, Score, transaction, validate)
+            Score,
+            self._response_for(
+                self._score_uri,
+                Score,
+                transaction,
+                validate,
+                hash_email,
+            ),
         )
 
     def report(self, report: Dict[str, Optional[str]], validate: bool = True) -> None:
@@ -547,15 +613,15 @@ class Client(BaseClient):
         :raises: AuthenticationError, InvalidRequestError, HTTPError,
           MinFraudError,
         """
-        prepared_request = self._prepare_report(report, validate)
+        prepared_request = prepare_report(report, validate)
         uri = self._report_uri
 
         response = self._do_request(uri, prepared_request)
         status = response.status_code
         content_type = response.headers["Content-Type"]
-        body = response.text
+        raw_body = response.text
         if status != 204:
-            raise self._exception_for_error(status, content_type, body, uri)
+            raise self._exception_for_error(status, content_type, raw_body, uri)
 
     def _response_for(
         self,
@@ -563,20 +629,23 @@ class Client(BaseClient):
         model_class: Union[Type[Factors], Type[Score], Type[Insights]],
         request: Dict[str, Any],
         validate: bool,
+        hash_email: bool,
     ) -> Union[Score, Factors, Insights]:
         """Send request and create response object."""
-        prepared_request = self._prepare_transaction(request, validate)
+        prepared_request = prepare_transaction(request, validate, hash_email)
 
         response = self._do_request(uri, prepared_request)
         status = response.status_code
         content_type = response.headers["Content-Type"]
-        body = response.text
+        raw_body = response.text
         if status != 200:
-            raise self._exception_for_error(status, content_type, body, uri)
-        return self._handle_success(body, uri, model_class)
+            raise self._exception_for_error(status, content_type, raw_body, uri)
+        return self._handle_success(raw_body, uri, model_class)
 
     def _do_request(self, uri: str, data: Dict[str, Any]) -> Response:
-        return self._session.post(uri, json=data, timeout=self._timeout)
+        return self._session.post(
+            uri, json=data, timeout=self._timeout, proxies=self._proxies
+        )
 
     def close(self):
         """Close underlying session
