@@ -16,9 +16,7 @@ class ValidationBase:
 
     def check_invalid_transaction(self, transaction):
         self.setup_transaction(transaction)
-        with self.assertRaises(
-            MultipleInvalid, msg="{0!s} is invalid".format(transaction)
-        ):
+        with self.assertRaises(MultipleInvalid, msg=f"{transaction} is invalid"):
             validate_transaction(transaction)
 
     def check_transaction(self, transaction):
@@ -26,7 +24,7 @@ class ValidationBase:
         try:
             validate_transaction(transaction)
         except MultipleInvalid as e:
-            self.fail("MultipleInvalid {0} thrown for {1}".format(e.msg, transaction))
+            self.fail(f"MultipleInvalid {e.msg} thrown for {transaction}")
 
     def check_transaction_str_type(self, object, key):
         self.check_transaction({object: {key: "string"}})
@@ -53,7 +51,7 @@ class ValidationBase:
 
     def check_invalid_report(self, report):
         self.setup_report(report)
-        with self.assertRaises(MultipleInvalid, msg="{0!s} is invalid".format(report)):
+        with self.assertRaises(MultipleInvalid, msg=f"{report} is invalid"):
             validate_report(report)
 
     def check_report(self, report):
@@ -61,13 +59,23 @@ class ValidationBase:
         try:
             validate_report(report)
         except MultipleInvalid as e:
-            self.fail("MultipleInvalid {0} thrown for {1}".format(e.msg, report))
+            self.fail(f"MultipleInvalid {e.msg} thrown for {report}")
 
     def check_report_str_type(self, key):
         self.check_report({key: "string"})
         self.check_invalid_report({key: 12})
 
 
+class TestTransaction(unittest.TestCase, ValidationBase):
+    def test_transaction_without_device(self):
+        transaction = {
+            "account": {
+                "user_id": "usr",
+            }
+        }
+        validate_transaction(transaction)
+
+
 class TestAccount(unittest.TestCase, ValidationBase):
     def test_account_user_id(self):
         self.check_transaction({"account": {"user_id": "usr"}})
@@ -134,17 +142,41 @@ class TestShippingAddress(unittest.TestCase, AddressBase):
 
 
 class TestCreditCard(ValidationBase, unittest.TestCase):
+    def test_country(self):
+        for code in ("CA", "US"):
+            self.check_transaction({"credit_card": {"country": code}})
+        for invalid in (1, None, "", "A1", "Canada"):
+            self.check_invalid_transaction({"credit_card": {"country": invalid}})
+
     def test_issuer_id_number(self):
-        for iin in ("123456", "532313"):
+        for iin in ("123456", "532313", "88888888"):
             self.check_transaction({"credit_card": {"issuer_id_number": iin}})
         for invalid in ("12345", "1234567", 123456, "12345a"):
             self.check_invalid_transaction(
                 {"credit_card": {"issuer_id_number": invalid}}
             )
 
+    def test_last_digits(self):
+        for last_digits in ("1234", "9323", "34"):
+            self.check_transaction({"credit_card": {"last_digits": last_digits}})
+        for invalid in ("12345", "123", 1234, "123a"):
+            self.check_invalid_transaction({"credit_card": {"last_digits": invalid}})
+        self.check_transaction(
+            {"credit_card": {"issuer_id_number": "88888888", "last_digits": "12"}}
+        )
+        self.check_transaction(
+            {"credit_card": {"issuer_id_number": "88888888", "last_digits": "1234"}}
+        )
+        self.check_transaction(
+            {"credit_card": {"issuer_id_number": "666666", "last_digits": "1234"}}
+        )
+        self.check_transaction(
+            {"credit_card": {"issuer_id_number": "666666", "last_digits": "34"}}
+        )
+
     def test_last_4_digits(self):
-        for iin in ("1234", "9323"):
-            self.check_transaction({"credit_card": {"last_4_digits": iin}})
+        for last_digits in ("1234", "9323", "34"):
+            self.check_transaction({"credit_card": {"last_4_digits": last_digits}})
         for invalid in ("12345", "123", 1234, "123a"):
             self.check_invalid_transaction({"credit_card": {"last_4_digits": invalid}})
 
@@ -177,6 +209,9 @@ class TestCreditCard(ValidationBase, unittest.TestCase):
         for invalid in ("\x20", "123456", "x" * 256):
             self.check_invalid_transaction({"credit_card": {"token": invalid}})
 
+    def test_was_3d_secure_successful(self):
+        self.check_bool("credit_card", "was_3d_secure_successful")
+
 
 class TestCustomInputs(ValidationBase, unittest.TestCase):
     def test_valid_inputs(self):
@@ -211,12 +246,17 @@ class TestDevice(ValidationBase, unittest.TestCase):
             self.check_invalid_transaction({"device": {"ip_address": invalid}})
 
     def test_missing_ip(self):
-        with self.assertRaises(MultipleInvalid):
-            validate_transaction({"device": {}})
+        validate_transaction({"device": {}})
+        validate_transaction(
+            {
+                "device": {
+                    "user_agent": "foo",
+                }
+            }
+        )
 
     def test_missing_device(self):
-        with self.assertRaises(MultipleInvalid):
-            validate_transaction({})
+        validate_transaction({})
 
     def test_user_agent(self):
         self.check_transaction_str_type("device", "user_agent")
